%
% $Id: heart-alsv.pl,v 1.50.2.3 2012-02-07 09:12:45 sbk Exp $
%
% Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
% HeaRT the ALSV(FD) handling
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- write('HeaRT: $Id: heart-alsv.pl,v 1.50.2.3 2012-02-07 09:12:45 sbk Exp $'),nl.

%%% Finite domain declaration
%%% domain(<name>,<list-of-values>).

%%% Set value definition
%%% set(<name>,<list-of-values>).

%%% Ordered Set value definition
%%% set(<name>,<list-of-values>,<order-predicate-name>).

%%% Attribute declaration
%%% attribute(<attribute-name>,<attribute-type>,<attribute-domain>).
% attribute(aDD,atomic,day_of_week).
% attribute(sDD,set,day_of_week).


%%% State definition: by listing all the values of the attributes
%%% state(<state-identifier>,<attribute>,<value>,<type>).
%%% state_time(<state-identifier>,<time>).
%%% <type> := atomic|set|...\
% state(s1,aMO,atomic,march).
% state(s2,sTM,set,[7,8,9,10,11,12,13,14,15]).


%%% Definition of facts: to be used in rule preconditions
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% <attribute-type> := atomic|set|....
%%% For <attribute-type>=atomic <relation> := \eq, \neq, \in, \not\in
%%% For <attribute-type>=set <relation> := \eq, \neq, \subseteq,\supseteq,\sim, \not\sim

%%% To be defined: 'NULL','_',[]


%%% Satisfaction of a single fact versus state: a fact 
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% is satisfied within some state <state> iff <relation> is valid within this state;
%%% Deatils are defined by the first rows of table1 and table2.
%%% valid(<fact>,<state>).


alsv_valid(Att eq any, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]),
	heart_debug(2,['Valid:',Att,eq,any]).

alsv_valid(Att in [any], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]),
	heart_debug(2,['Valid:',Att,in,[any]]).
	
alsv_valid(Att eq any, State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	heart_debug(2,['Valid:',Att,eq,any]).

alsv_valid(Att eq [any], State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	heart_debug(2,['Valid:',Att,eq,[any]]).

alsv_valid(Att eq null, State) :-
	alsv_values_get(State,Att,StateValue),
	!,
	\+ alsv_values_get(State,Att,StateValue), 
	heart_debug(2,['Valid:',Att,eq,null]).
	
alsv_valid(Att eq null, State) :-
	alsv_values_get(State,Att,StateValue),
	!,
	StateValue = null, 					   
	heart_debug(2,['Valid:',Att,eq,null]).

alsv_valid(Att in [L to U], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,[L],[NormL]),
	normalize(Att,[U],[NormU]),
	NormStateValue =< NormU,
	NormStateValue >= NormL,		
	heart_debug(2,['Valid:',Att,in,[L to U]]).
	
alsv_valid(Att notin [L to U], State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,[L],[NormL]),
	normalize(Att,[U],[NormU]),
	NormStateValue >= NormU,
	NormStateValue =< NormL,	
	heart_debug(2,['Valid:',Att,notin,[L to U]]).
	
alsv_valid(Att in SetValue, State) :-		
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), 	
	alsv_values_check(Att,SetValue),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,SetValue,NormSetValue),
	(member(NormStateValue,NormSetValue);math_member(NormStateValue,NormSetValue)),		
	heart_debug(2,['Valid:',Att,in,SetValue]).
	
alsv_valid(Att notin SetValue, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,SetValue),
	normalize(Att,[StateValue],[NormStateValue]),
	normalize(Att,SetValue,NormSetValue),
	\+ member(NormStateValue,NormSetValue),
	\+ math_member(NormStateValue,NormSetValue),
	heart_debug(2,['Valid:',Att,notin,SetValue]).
	
alsv_valid(Att eq Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue =:= NormStateValue,
	heart_debug(2,['Valid:',Att,eq,Value]).
	
alsv_valid(Att eq Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue == NormStateValue,
	heart_debug(2,['Valid:',Att,eq,Value]).
	
alsv_valid(Att neq Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue \= NormStateValue,
	heart_debug(2,['Valid:',Att,neq,Value]).

alsv_valid(Att gt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),	
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue < NormStateValue,
	heart_debug(2,['Valid:',Att,gt,Value]).
	
alsv_valid(Att gte Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),	
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue =< NormStateValue,
	heart_debug(2,['Valid:',Att,gte,Value]).
	
alsv_valid(Att lt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue > NormStateValue,
	heart_debug(2,['Valid:',Att,lt,Value]).
	
alsv_valid(Att lte Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	normalize(Att,[Value],[NormValue]),
	normalize(Att,[StateValue],[NormStateValue]),
	NormValue >= NormStateValue,
	heart_debug(2,['Valid:',Att,lte,Value]).
	
%%% Table2: cases of \eq, \neq, \subseteq, \supseteq,\sim, \notsim  versus set StateValue in State
alsv_valid(Att eq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	eqset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att neq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	neqset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att subseteq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	alsv_subset(NormStateValue,NormSet),
	heart_debug(2,['Valid:',Att,subset,Set]).
	
alsv_valid(Att supseteq Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	alsv_subset(NormSet,NormStateValue),
	heart_debug(2,['Valid:',Att,supset,Set]).
	
alsv_valid(Att sim Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	simset(NormStateValue,NormSet),		
	heart_debug(2,['Valid:',Att,sim,Set]).
	
alsv_valid(Att notsim Set,State) :-
	alsv_attr_class(Att,general),
	alsv_values_get(State,Att,StateValue),
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	normalize(Att,StateValue,NormStateValue),
	normalize(Att,Set,NormSet),
	\+simset(NormStateValue,NormSet),		
	heart_debug(2,['Valid:',Att,notsim,Set]).

alsv_valid(F,_) :-
	heart_debug(2,['Check for:',F,'NYI!']),
	fail.
	
alsv_values_get(State,Att,StateValue) :-
	xstat State: [Att,StateValue],!.
	
alsv_values_get(_,Att,StateValue) :-
	xattr(A),
	member(name:Att,A),
	member(comm:C,A),
	(C=in;C=inter),
	fire_callback(Att),			
	xstat current: [Att,StateValue].
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% alsvd checks
alsv_values_check(_,[]).

	
%domain is set, but value X to Y
alsv_values_check(Att,[LS to US|Rest]) :-
	alsv_domain(Att,Domain,numeric), 
	heart_debug(2,['Check:',Att,numericreg,LS to US]),		
	math_member([LS to US],Domain),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),	
	alsv_domain(Att,Domain,numeric),
	heart_debug(2,['Check:',Att,numeric,StateValue]),	
	math_member(StateValue,Domain), % check if value in the domain with math_memeber
	alsv_values_check(Att,Rest),
	!.
	
%domain is set, but value X to Y
alsv_values_check(Att,[LS to US|Rest]) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,symbolic), 
	heart_debug(2,['Check:',Att,symbolic,LS to US]),		
	normalize(Att,Domain,NDomain),	
	normalize(Att,LS,NLS),
	normalize(Att,US,NUS),
	sort(NDomain, [First|SNDomain]),
	last(SNDomain,Last),
	NLS >= First,
	NUS =< Last,
	alsv_values_check(Att,Rest),
	!.
		
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	number(StateValue),
	math_member(StateValue,Domain), %chceck if weight in the domain with math_memeber
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	member(StateValue/_,Domain),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	number(StateValue),
	truncate(StateValue) =:= StateValue,
	TruncatedValue is truncate(StateValue),
	nth1(TruncatedValue,Domain,_),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	member(StateValue,Domain),
	alsv_values_check(Att,Rest),
	!.

% alsv_values_check for general atts missing! ex. [1,2,3,4,5] domain [1,2,3,4,5,6,7,8]. subset

alsv_values_check(Att,[V|_]) :-
	write('  ERROR: values of \"'), write(Att),
	write('\": '), write(V), write(' NOT in the domain!'),nl,!.

%retrieves attribute domain, and base: symbolic, numeric
%Does not scale nor normalize domains that are for instance january/1, february/2 
alsv_domain(Att,Domain,Base) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(domain:Domain,Ts),
	member(base:Base,Ts).

alsv_normalized_domain(Att,Domain,Base) :-
   alsv_domain(Att,RawDomain,Base),
   normalize_all(Att,RawDomain,Domain).

% retrieves attribute class: simple, general
alsv_attr_class(Att,Class) :-
	xattr(A),
	member(name:Att,A),
	member(class:Class,A).

alsv_order(Att,yes) :-
	xattr(A),
	member(name:Att,A),
	member(type:T	,A),
	xtype(Ts),	
	member(name:T,Ts),
	member(base:numeric, Ts),!.

alsv_order(Att,Order) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(ordered:Order,Ts),!.

alsv_order(_,no).
	

% heart
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

alsv_transition(Att set Expr) :-
%	atom(Val),
	\+ Expr = [_|_],
	alsv_domain(Att,_,numeric),
	alsv_attr_class(Att,simple),
	alsv_eval_simple(Expr, Val),
	heart_debug(2,['Adding new fact: ',Att,' set to ',Val]),
	alsv_new_val(Att,Val).

alsv_transition(Att set Val) :-
	\+ Val = [_|_],
	alsv_domain(Att,_,symbolic),
	alsv_attr_class(Att,simple),
	heart_debug(2,['Adding new fact: ',Att,' set to ',Val]),
	alsv_new_val(Att,Val).

alsv_transition(Att set Expr) :-
	Expr = [_|_],
	alsv_eval_general(Expr,Val),
	alsv_attr_class(Att,general),
	alsv_new_val(Att,Val).

% fixme: no direct assert/retract, but set/get_att_val
alsv_new_val(A,V):-
	retractall(xstat current:[A,_]),
	assert(xstat current:[A,V]),
	heart_debug(2,['Trans:',A,to,V]).

alsv_eval_simple(Expr, Evaluated) :-
	parse(Expr, Parsed),
	Resolved =.. Parsed,
	Evaluated is Resolved.

alsv_eval_general(Expr, Expr) :-
	is_list(Expr),!.

alsv_eval_general(Expr, Eval) :-
	Expr = union(V1,V2),
	alsv_eval_general(V1,R1),
	alsv_eval_general(V2,R2),
	alsv_union(R1,R2,Eval),!.

alsv_eval_general(Expr, Eval) :-
	Expr = intersect(V1,V2),
	alsv_eval_general(V1,R1),
	alsv_eval_general(V2,R2),
	alsv_intersect(R1,R2,Eval),!.

alsv_eval_general(Expr, Eval) :-
	Expr = except(V1,V2),
	alsv_eval_general(V1,R1),
	alsv_eval_general(V2,R2),
	alsv_except(R1,R2,Eval),!.

alsv_eval_general(Expr, Eval) :-
	Expr = complement(Att,V2),
	alsv_eval_general(V2,R2),
	alsv_complement(Att,R2,Eval),!.

parse(Expr, Parsed) :-
	Expr =.. List,
	parse_helper(List,Parsed),!.	

parse_helper([],[]).

parse_helper([H|Rest],[Val|Parsed]) :-
	parse_is_attr(H),
	alsv_values_get(current,H,Val),
	parse_helper(Rest,Parsed),!.
	
parse_helper([H|Rest],[H|Parsed]) :-
	number(H),
	parse_helper(Rest, Parsed),!.	

parse_helper([H|Rest],[H|Parsed]) :-
	parse_is_operator(H),
	parse_helper(Rest,Parsed),!.

parse_helper([SubExpr|Rest], [Unsplit|Parsed]) :-
	SubExpr =.. Split,
	parse_helper(Split,Subparsed),
	Unsplit =.. Subparsed,
	parse_helper(Rest, Parsed),!.


parse_is_operator(H) :-
	H = '-';
	H = '+';
	H = '*';
	H = '/';
	H = ')';
	H = ')'.

parse_is_attr(H) :-
	xattr(P),
	member(name:H,P),!.
	


