%
% $Id: heart-rules.pl,v 1.13.4.2 2012-02-07 09:12:45 sbk Exp $
%
%     Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
%     HeaRT the main shell
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2009   by S. Bobek, M. Gawedzki
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- write('HeaRT: $Id: heart-rules.pl,v 1.13.4.2 2012-02-07 09:12:45 sbk Exp $'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Rule in table processing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% rule processing
cond_satisfy([]).
cond_satisfy([C|Rest]) :-
	formula_satisfied(C),!,
	cond_satisfy(Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FIXME
% here comez the alsv magic
% fill all the clauses implementing different operators
% for both simple and general attributes
%formula_satisfied(C) :-
%	write('  *Condition: '), write(C),
%	write(' satisfied.'), nl.
formula_satisfied(C) :-
	alsv_valid(C,current),!,
	heart_debug(1,['Condition: ',C,' satisfied']).

	%DEBUG OBLY
formula_satisfied(C) :-
	heart_debug(1,['Condition: ',C,' NOT satisfied']),!,fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


inference_transfer(T/R) :-
	give_token(T),!,
	heart_debug(1,['Go to:',T/R]).
inference_transfer(T) :-
	give_token(T),!,
	heart_debug(1,['Go to:',T]).
inference_transfer(none) :-
	heart_debug(1,['Go to:','NONE']),		
	!.

give_token(Table) :-
	xtoken(Table,Get,Req),
	Get < Req,
	retractall(xtoken(Table,_,_)),
	IncGet is Get+1,
	assert(xtoken(Table,IncGet,Req)),!.
give_token(_).

process_decision(D**>A:G,G) :-
	process_dec_attributes(D),
	process_dec_actions(A),
	!.
process_decision(D**>A,none) :-
	process_dec_attributes(D),
	process_dec_actions(A),
	!.
process_decision(D:Go,Go) :-
	process_dec_attributes(D),
	!.
process_decision(D,none) :-
	process_dec_attributes(D),
	!.

process_dec_attributes([]).
process_dec_attributes([At|Rest]):-
	process_decision(At),
	process_dec_attributes(Rest).

process_dec_actions([]).
process_dec_actions([Ac|Rest]):-
	process_action(Ac),
	process_dec_actions(Rest).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FIXME
% here comez the decision attribute manipulation magic
% fill all the clauses implementing different cases
% including set, setheoretic, arithm expressions
% for both simple and general attributes
% this needs to address some middle layer to handle blackboard
process_decision(Att) :-
	alsv_transition(Att).

process_decision(Att) :-
	heart_debug(1,['Decision act: ',Att,'NOT handled!']).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FIXME
% here comez the decision action execution magic
% fill all the clauses implementing different cases
% executing preregistered predicates, prolog predicates, java?
% this needs to address some middle layer to handle blackboard
process_action([ActionName, Parameters]) :-
	xactn ActionName : Parameters >>> Goal,
	call(Goal),
	heart_debug(1,['Decision act: ',ActionName,'processes.']).	
	% check if it is a registered action
	% check if arguments are OK
	% call(Ac), % hopefully it would not expolode?

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% init+debug helper stuff
find_free_atts(FreeAtts) :-
	findall(D,xtable(_,_,_,D),Decs),
	findall(C,xtable(_,_,C,_),Conds),	
	flatten(Decs,FD),
	flatten(Conds,FC),
	subtract(FC,FD,FreeAtts).

init_tables(RTables) :-
	find_free_atts(FA),
	findall(I,list_initabs(FA,I),RTables).
list_initabs(FA,I) :-	
	xtable(I,_,CN,_),
	subset(CN,FA).

xtables_show :-
	xschm S: _ ==> _,
	xtable_show(S),
	fail.
xtables_show.

xtable_show(T) :-
	xrule T/R: _ ==> _,
	write(T/R), nl,
	fail.
xtable_show.

xtable(Name,Desc,From,To) :-
	xschm Name/Desc : From ==> To.
xtable(RealName,Desc,From,To) :-
	xschm RealName : From ==> To,
	RealName \= _/Desc.


heart_debug(N,D) :-
	flag_debug(L),
	number(L),
	N =< L,
	write('DEBUG('),
	write(N), write('): '),
	heart_debug_process(D),!.
heart_debug(_,_).

heart_debug_process([]) :- nl,!.
heart_debug_process([H|T]):-
	write(H), write(' '),!,
	heart_debug_process(T).

