%
% $Id: heart.pl,v 1.23.2.2 2012-02-07 09:12:45 sbk Exp $
%
%     Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
%     HeaRT the main shell
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2009   by S. Bobek, M. Gawedzki
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- write('HeaRT: $Id: heart.pl,v 1.23.2.2 2012-02-07 09:12:45 sbk Exp $'),nl.
:- write('\nCopyright (C) by G. J. Nalepa and A. Ligeza 2004-2006 and the HeKatE Project 2006-2009\nHeaRT is free software: you can redistribute it and/or modify it under\nthe terms of the GNU General Public License as published by the Free Software Foundation,\neither version 3 of the License, or (at your option) any later version.\n\n').
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Includes
:- ensure_loaded('config.pl').
:- ensure_loaded('util.pl').
:- ensure_loaded('heart-alsv.pl').
:- ensure_loaded('heart-infer.pl').
:- ensure_loaded('heart-rules.pl').
:- ensure_loaded('heart-state.pl').
:- ensure_loaded('heart-io.pl').
:- ensure_loaded('verification/verification.pl').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Help
%
hlp :-
	nl,
	write('To load an example, simply put [therm-rt].'), nl, nl,
	write('Inference:'),nl,
	write('  gox(+State, +Tables, +Mode). - start from given state,'),nl,
	write('              inference based on explicit Tables list in a given Mode (foi,ddi,gdi,tdi)'),nl,
	write('  goxio(+Tables, +Mode). - like above, but ask user for the input state.'),nl,nl,
	write('Verification (HalVA):'),nl,	
	write('  check_grammar(+File).'),nl,
	write('  vsubsume(+Schema).'),nl,
	write('  vcomplete(+Schema).'),nl,
	write('  vcontrdict(+Schema).'),nl,
	write('  vreduce(+Schema).'),nl,
	write('  io_export_TEX(+Filename).'),nl,nl,
	write('Run server'),nl,
	write('  srv. - run server on the default port'),nl,
	write('  srv(+Port). - run server on a given port'),nl,nl,
	write('hlt. - goodbye.'), nl,
	nl.
:-
	nl,
	write('This is HeaRT, the HeKatE Run Time engine.'), nl,
	write('Use \'hlp.\' to get some help'), nl,
	nl.

hlt :- halt.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main shell predicates

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% gox in det.
%
%  Runs the Heart XTT interpreter


gox(StateIn,T,foi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runtables(T),
	trajectory_projection(end,simple), 
	state_print(current),!.

gox(StateIn,T,ddi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_DDI(T),
	trajectory_projection(end,simple), 
	state_print(current),!.

gox(StateIn,T,tdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_TDI(T),
	trajectory_projection(end,simple), 
	state_print(current),!.
	
gox(StateIn,T,gdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_GDI(T),
	trajectory_projection(end,simple), 
	state_print(current),!.

goxio(StateIn,T,foi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runtables(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(StateIn,T,ddi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_DDI(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(StateIn,T,tdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_TDI(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(StateIn,T,gdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	state_print(StateIn),
	state_copy(StateIn,current),
	trajectory_projection(start,_), 
	runrules_GDI(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.

goxio(T,foi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runtables(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,ddi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_DDI(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,tdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_TDI(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.
	
goxio(T,gdi) :-
	traj_genid(_),
	% check/bootstrap attributes
	% built a set of runable tables
	%findall_input_atts(IA), %all atts having 'comm: in'
	%fireall_callbacks(IA),  %fire all callbacks in given atts
	!,
	trajectory_projection(start,_), 
	runrules_GDI(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	trajectory_projection(end,simple), 
	state_print(current),!.


	
table_print :-
	write('Tables: '),
	table_print_helper.
table_print_helper :-		
	xschm T:_==>_,
	write(T),write(', '),
	fail.
table_print_helper.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% knowledge base syntax and structure checks
%
xtype_def([
	   [name, [_]],
	   [base, [symbolic,numeric]],
	   [ordered,[yes,no]],
	   [domain,[_]],
	   %[abbrev,[_]],
	   [desc, [_]],
	   [numbering,[_]],
	   [length, [_]],
	   [scale,[_]]	  
	  ]
	  ).

xtype_check :-
	xtype(T),
	member(name:N,T),
	write('Checking xtype: "'), write(N), write('"  '), nl,
	xtype_check_single(T),
	nl,
	fail.
xtype_check :-
	write('No more xtypes to check.').

xtype_check_single(T) :-
	member(D:V,T),
	xtype_def(Y),
	member([D,_],Y),
	write('    '),write(D:V), write(' -> OK').

xattr_def([
	   [name,[_]],
	   [class,[simple,general]],
	   [type,[_]],
	   [comm,[in,inter,out,comm]],
	   [abbrev,[_]],
	   [desc,[_]],
	   [callback,[_]]
	  ]).

%xrules_check :-.
% scheme exists
% attributes valid with scheme
% attributes defined
% on the runtimelevel: see if val is in the att domain!
% if control spec exists, check table and rule id
