:- dynamic (xtype)/1.

xtype[name:boolean, base:numeric, length:1, scale:0, desc:boolean, domain:[0, 1]].
xtype[name:integer, base:numeric, length:10, scale:0, desc:integer, domain:[-2147483648 to 2147483647]].
xtype[name:default, base:symbolic, domain:[none/1], ordered:yes].
xtype[name:tPin, base:numeric, length:4, scale:0, desc:'Represents the PIN numbers ', domain:[0 to 9999]].
xtype[name:tPinDifference, base:numeric, length:4, scale:0, desc:'Represents all the possible results of two PIN numbers subtraction ', domain:[-9999 to 9999]].
xtype[name:tUserActions, base:symbolic, desc:'The set of actions that user can invoke', domain:[withdraw, balance]].
xtype[name:tCashPointActivities, base:symbolic, desc:'The set of actions that can be executed by cashpoint', domain:[askForPIN, payOut, takeCardAway, displayBalance, msgNotEnoughFoundsInMachine, msgNotEnoughFoundsOnAccount]].
xtype[name:tMoney, base:numeric, domain:[0 to 1000000], desc:'Represents the amount of money'].
xtype[name:tBool, base:symbolic, desc:'The boolean values', domain:[false/0, true/1], ordered:yes].
xtype[name:tAttempts, base:numeric, length:1, scale:0, desc:' yes \t The type that defines the number of attempts to enter a correct PIN', domain:[0 to 3]].

:- dynamic (xtpgr)/1.


:- dynamic (xattr)/1.

xattr[name:authorizated, abbrev:autho, class:simple, type:tBool, comm:inter].
xattr[name:cashPointActivity, abbrev:cashP, class:simple, type:tCashPointActivities, comm:out, callback:[xpce_say_symbolic, [cashPointActivity]]].
xattr[name:cashPointAmount, abbrev:cash, class:simple, type:tMoney, comm: (in), callback:[xpce_ask_numeric, [cashPointAmount]]].
xattr[name:cdAmountDifference, abbrev:cdAmo, class:simple, type:tMoney, comm:inter].
xattr[name:correctPin, abbrev:corre, class:simple, type:tPin, comm: (in), callback:[xpce_ask_numeric, [correctPin]]].
xattr[name:desiredAmount, abbrev:desir, class:simple, type:tMoney, comm: (in), callback:[xpce_ask_numeric, [desiredAmount]]].
xattr[name:enteredPin, abbrev:enter, class:simple, type:tPin, comm: (in), callback:[xpce_ask_numeric, [enteredPin]]].
xattr[name:failedAttempts, abbrev:faile, class:simple, type:tAttempts, comm:inter].
xattr[name:pinDifference, abbrev:pinDi, class:simple, type:tPinDifference, comm:inter].
xattr[name:udAmountDifference, abbrev:udAmo, class:simple, type:tMoney, comm:inter].
xattr[name:userAccountAmount, abbrev:userA, class:simple, type:tMoney, comm: (in), callback:[xpce_ask_numeric, [userAccountAmount]]].
xattr[name:userRequestedAction, abbrev:userR, class:simple, type:tUserActions, comm: (in), callback:[xpce_ask_symbolic, [userRequestedAction]]].

:- dynamic (xatgr)/1.


:- dynamic (xschm)/1.

xschm cashPointActivityDetermining:[authorizated, failedAttempts, userRequestedAction, udAmountDifference, cdAmountDifference]==>[cashPointActivity].
xschm authorization:[pinDifference]==>[authorizated, failedAttempts].
xschm pinDifferenceCounting:[enteredPin, correctPin]==>[pinDifference].
xschm udAmountDifferenceCounting:[desiredAmount, userAccountAmount]==>[udAmountDifference].
xschm cdAmountDifferenceCounting:[cashPointAmount, desiredAmount]==>[cdAmountDifference].

:- dynamic (xrule)/1.

xrule cashPointActivityDetermining/1:[authorizated eq false, failedAttempts lt 3, userRequestedAction eq any, udAmountDifference eq any, cdAmountDifference eq any]==>[cashPointActivity set askForPIN].
xrule cashPointActivityDetermining/2:[authorizated eq false, failedAttempts eq 3, userRequestedAction eq any, udAmountDifference eq any, cdAmountDifference eq any]==>[cashPointActivity set takeCardAway].
xrule cashPointActivityDetermining/3:[authorizated eq true, failedAttempts eq any, userRequestedAction eq balance, udAmountDifference eq any, cdAmountDifference eq any]==>[cashPointActivity set displayBalance].
xrule cashPointActivityDetermining/4:[authorizated eq true, failedAttempts eq any, userRequestedAction eq withdraw, udAmountDifference gte 0.0, cdAmountDifference gte 0.0]==>[cashPointActivity set payOut].
xrule cashPointActivityDetermining/5:[authorizated eq true, failedAttempts eq any, userRequestedAction eq withdraw, udAmountDifference eq any, cdAmountDifference lt 0.0]==>[cashPointActivity set msgNotEnoughFoundsInMachine].
xrule cashPointActivityDetermining/6:[authorizated eq true, failedAttempts eq any, userRequestedAction eq withdraw, udAmountDifference lt 0.0, cdAmountDifference gte 0.0]==>[cashPointActivity set msgNotEnoughFoundsOnAccount].
xrule authorization/1:[pinDifference neq 0]==>[authorizated set false, failedAttempts set (failedAttempts+1)]:cashPointActivityDetermining.
xrule authorization/2:[pinDifference eq 0]==>[authorizated set true, failedAttempts set failedAttempts].
xrule pinDifferenceCounting/1:[enteredPin eq any, correctPin eq any]==>[pinDifference set (correctPin-enteredPin)]:authorization.
xrule udAmountDifferenceCounting/1:[desiredAmount eq any, userAccountAmount eq any]==>[udAmountDifference set (userAccountAmount-desiredAmount)]:cashPointActivityDetermining.
xrule cdAmountDifferenceCounting/1:[cashPointAmount eq any, desiredAmount eq any]==>[cdAmountDifference set (cashPointAmount-desiredAmount)]:cashPointActivityDetermining.

:- dynamic (xstat)/1.

xstat init0:[enteredPin, 1111].
xstat init0:[correctPin, 1111].
xstat init0:[desiredAmount, 100.0].
xstat init0:[cashPointAmount, 100000.0].
xstat init0:[userAccountAmount, 10000.0].
xstat init0:[userRequestedAction, withdraw].
xstat init0:[failedAttempts, 0].
xstat init1:[correctPin, 1112].
xstat init1:[cashPointAmount, 100000.0].
xstat init1:[userAccountAmount, 10000.0].
xstat init1:[failedAttempts, 0].
xstat init2:[enteredPin, 1111].
xstat init2:[correctPin, 1111].
xstat init2:[desiredAmount, 100.0].
xstat init2:[cashPointAmount, 10.0].
xstat init2:[userAccountAmount, 10000.0].
xstat init2:[userRequestedAction, withdraw].
xstat init2:[failedAttempts, 0].
xstat init3:[enteredPin, 1111].
xstat init3:[correctPin, 1111].
xstat init3:[desiredAmount, 100.0].
xstat init3:[cashPointAmount, 1000.0].
xstat init3:[userAccountAmount, 10.0].
xstat init3:[userRequestedAction, withdraw].
xstat init3:[failedAttempts, 0].
xstat init4:[enteredPin, 1111].
xstat init4:[correctPin, 1112].
xstat init4:[failedAttempts, 2].
xstat init4:[userRequestedAction, withdraw].
xstat init4:[failedAttempts, 0].
xstat init5:[enteredPin, 1111].
xstat init5:[correctPin, 1111].
xstat init5:[desiredAmount, 100.0].
xstat init5:[cashPointAmount, 100000.0].
xstat init5:[userAccountAmount, 10000.0].
xstat init5:[userRequestedAction, balance].
xstat init5:[failedAttempts, 0].
xstat init6:[enteredPin, 1111].
xstat init6:[correctPin, 1112].
xstat init6:[desiredAmount, 100.0].
xstat init6:[cashPointAmount, 100000.0].
xstat init6:[userAccountAmount, 10000.0].
xstat init6:[userRequestedAction, balance].
xstat init6:[failedAttempts, 0].
xstat init7:[enteredPin, 1111].
xstat init7:[correctPin, 1111].
xstat init7:[desiredAmount, 100.0].
xstat init7:[cashPointAmount, 10.0].
xstat init7:[userAccountAmount, 10000.0].
xstat init7:[userRequestedAction, balance].
xstat init7:[failedAttempts, 0].
xstat init8:[enteredPin, 1111].
xstat init8:[correctPin, 1111].
xstat init8:[desiredAmount, 100.0].
xstat init8:[cashPointAmount, 1000.0].
xstat init8:[userAccountAmount, 10.0].
xstat init8:[userRequestedAction, balance].
xstat init8:[failedAttempts, 0].
xstat init9:[enteredPin, 1111].
xstat init9:[correctPin, 1112].
xstat init9:[failedAttempts, 2].
xstat init9:[userRequestedAction, balance].
xstat init9:[failedAttempts, 0].
xstat s1:[enteredPin, 1111].
xstat s1:[correctPin, 1111].
xstat s1:[desiredAmount, 100.0].
xstat s1:[cashPointAmount, 100000.0].
xstat s1:[userAccountAmount, 10000.0].
xstat s1:[userRequestedAction, withdraw].
xstat s1:[failedAttempts, 0].
xstat s2:[enteredPin, 1111].
xstat s2:[correctPin, 1111].
xstat s2:[desiredAmount, 100.0].
xstat s2:[cashPointAmount, 100000.0].
xstat s2:[userAccountAmount, 10000.0].
xstat s2:[userRequestedAction, withdraw].
xstat s2:[failedAttempts, 0].
xstat current:[enteredPin, 1111].
xstat current:[correctPin, 1111].
xstat current:[desiredAmount, 100.0].
xstat current:[cashPointAmount, 100000.0].
xstat current:[userAccountAmount, 10000.0].
xstat current:[userRequestedAction, withdraw].
xstat s3:[enteredPin, 1111].
xstat s3:[correctPin, 1111].
xstat s3:[desiredAmount, 100.0].
xstat s3:[cashPointAmount, 100000.0].
xstat s3:[userAccountAmount, 10000.0].
xstat s3:[userRequestedAction, withdraw].
xstat s3:[failedAttempts, 0].
xstat current:[pinDifference, 0].
xstat s4:[enteredPin, 1111].
xstat s4:[correctPin, 1111].
xstat s4:[desiredAmount, 100.0].
xstat s4:[cashPointAmount, 100000.0].
xstat s4:[userAccountAmount, 10000.0].
xstat s4:[userRequestedAction, withdraw].
xstat s4:[failedAttempts, 0].
xstat s4:[pinDifference, 0].
xstat current:[authorizated, true].
xstat current:[failedAttempts, 0].
xstat s5:[enteredPin, 1111].
xstat s5:[correctPin, 1111].
xstat s5:[desiredAmount, 100.0].
xstat s5:[cashPointAmount, 100000.0].
xstat s5:[userAccountAmount, 10000.0].
xstat s5:[userRequestedAction, withdraw].
xstat s5:[pinDifference, 0].
xstat s5:[authorizated, true].
xstat s5:[failedAttempts, 0].
xstat current:[udAmountDifference, 9900.0].
xstat s6:[enteredPin, 1111].
xstat s6:[correctPin, 1111].
xstat s6:[desiredAmount, 100.0].
xstat s6:[cashPointAmount, 100000.0].
xstat s6:[userAccountAmount, 10000.0].
xstat s6:[userRequestedAction, withdraw].
xstat s6:[pinDifference, 0].
xstat s6:[authorizated, true].
xstat s6:[failedAttempts, 0].
xstat s6:[udAmountDifference, 9900.0].
xstat current:[cdAmountDifference, 99900.0].
xstat s7:[enteredPin, 1111].
xstat s7:[correctPin, 1111].
xstat s7:[desiredAmount, 100.0].
xstat s7:[cashPointAmount, 100000.0].
xstat s7:[userAccountAmount, 10000.0].
xstat s7:[userRequestedAction, withdraw].
xstat s7:[pinDifference, 0].
xstat s7:[authorizated, true].
xstat s7:[failedAttempts, 0].
xstat s7:[udAmountDifference, 9900.0].
xstat s7:[cdAmountDifference, 99900.0].
xstat current:[cashPointActivity, payOut].
xstat s8:[enteredPin, 1111].
xstat s8:[correctPin, 1111].
xstat s8:[desiredAmount, 100.0].
xstat s8:[cashPointAmount, 100000.0].
xstat s8:[userAccountAmount, 10000.0].
xstat s8:[userRequestedAction, withdraw].
xstat s8:[pinDifference, 0].
xstat s8:[authorizated, true].
xstat s8:[failedAttempts, 0].
xstat s8:[udAmountDifference, 9900.0].
xstat s8:[cdAmountDifference, 99900.0].
xstat s8:[cashPointActivity, payOut].

:- dynamic (xhalv)/1.


:- dynamic (xtraj)/1.

xtraj t1:[[start, s1]].
xtraj t2:[[start, s2]].
xtraj t3:[[start, s3], [pinDifferenceCounting/1, s4], [authorization/2, s5], [udAmountDifferenceCounting/1, s6], [cdAmountDifferenceCounting/1, s7], [cashPointActivityDetermining/4, s8]].

