%debug_set_flag(+Val)
%Predicat that sets the debug level.
%It takes one parameter that is the integer value from 0 to 3, where 0 means no details, and 3 means all details.
debug_set_flag(Val) :-
	retractall(flag_debug(_)),
	assert(flag_debug(Val)).


%get_explicit_value(+Att,+NormValue,?ExplicitValue)
%In case an attiribute is ordered it is stored in HeaRT memory in its numerical representation
%The predicate is used to retriev its symbolic value (for instance january instead of 1.0)
get_explicit_value(Att,[NLB to NUB],[LB to UB]) :-
	get_explicit_value(Att,NLB,LB),
	get_explicit_value(Att,NUB,UB).

get_explicit_value(Att,NormValue,ExplicitValue) :-
	alsv_normalized_domain(Att,NormDomain,symbolic),
	alsv_domain(Att,Domain,_),
	nth0(Index,NormDomain,NormValue),
	nth0(Index,Domain,ExplicitValue),!.
	
%For states gives as a single values but still in brackets, eg [day,[6.0]]
get_explicit_value(Att,[NormValue],ExplicitValue) :-
	alsv_normalized_domain(Att,NormDomain,symbolic),
	alsv_domain(Att,Domain,_),
	nth0(Index,NormDomain,NormValue),
	nth0(Index,Domain,ExplicitValue),!.
	
get_explicit_value(Att,NormValue,NormValue) :-
	alsv_domain(Att,_,numeric).


%%%%%%%%%%%%%%%%%%% ALSV(FD) UTILITIES %%%%%%%%%%%%%%%%%%%%%%%%%%%%
normalize_all(_,[],[]).
normalize_all(Att,[Value|RawRest],[Output|Rest]) :-
     	normalize(Att,[Value],[Output]),
        normalize_all(Att,RawRest,Rest).

	
normalize(Att, Value, Output) :-
	\+ alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(Value,Domain),
	Value = Output,!.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	number(Value),			%and change output to floats if number (can be outside the domain)
	alsv_domain(Att,Domain,_),
	math_member(Value,Domain),	
	Output is Value*1.0,!.

normalize(Att, Value, Output) :-	
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(Value/Numbering,Domain), 
	Output is Numbering*1.0,!.

normalize(Att, MixedValue, Output) :-	
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	member(MixedValue,Domain),
	MixedValue = _/Numbering, 
	Output is Numbering*1.0,!.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	atom(Value),			%and change output to floats if number
	nth1(Number,Domain,Value),
	Output is Number*1.0.
	
normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	alsv_domain(Att,Domain,_),
	number(Value),
	truncate(Value) =:= Value,
	TruncatedNumber is truncate(Value),
	nth1(TruncatedNumber,Domain,_),	%and change output to floats
	Output is TruncatedNumber*1.0,!.

normalize(Att,[X to Y | Rest1],Rest2) :-
	%alsv_domain(Att,_,numeric),	
	normalize(Att,X,XN),
	normalize(Att,Y,YN),
	normalize(Att,Rest1,Rest),
	merge([XN to YN],Rest,Rest2),
	!.
	

normalize(Att,[Number|Rest1],[Value|Rest2]) :-
	normalize(Att,Number,Value),
	normalize(Att,Rest1,Rest2),
	!.
	
normalize(_,[],[]).
		

math_member(Number, [Lower to Upper|_]) :-
	number(Number),
	number(Lower),
	number(Upper),
	Number =< Upper,
	Number >= Lower,!.


math_member(Number, [Lower to Upper|Rest]) :-
	number(Number),
	number(Lower),
	number(Upper),
	math_member(Number,Rest),!.

math_member(L to U, Set) :-
	is_list(Set),
	member(L,Set),
	member(U,Set),!.

math_member(L to U,[Number|Rest]) :-
	number(Number),
	math_member(L to U,Rest),!.

math_member(L to U,[SL to SU|_]) :-
	number(L),
	number(U),
	number(SL),
	number(SU),
	L >= SL,
	U =< SU,!.

math_member(L to U,[SL to SU|Rest]) :-
	number(L),
	number(U),
	number(SL),
	number(SU),
	math_member(L to U,Rest),!.


math_member(Weight,[_/DomainWeight|[]]) :-
	number(Weight),
	number(DomainWeight),
	Weight =:= DomainWeight,!. 

math_member(Weight,[_/DomainWeight|Rest]) :-
	number(Weight),
	number(DomainWeight),
	Weight =\= DomainWeight, 
	math_member(Weight,Rest),!.
	
math_member(Weight,[_/DomainWeight|_]) :-
	number(Weight),
	number(DomainWeight),
	Weight =:= DomainWeight,!.
	
math_member(Number,[Head|[]]) :-
	number(Number),
	number(Head),
	Number =:= Head,!. 
math_member(Number,[Head|Rest]) :-
	number(Number),
	number(Head),
	Number =\= Head, 
	math_member(Number,Rest),!.
	
math_member(Number,[Head|_]) :-
	number(Number),
	number(Head),
	Number =:= Head,!.


math_member([],_).

math_member([Head|Rest],Set) :-
	math_member(Head,Set),
	math_member(Rest,Set),!.

alsv_subset(Set1, Set2) :-
	subset(Set1,Set2),!.
alsv_subset(Set1,Set2) :-
	math_member(Set1,Set2),!.

eqset(Set1,Set2) :-
	alsv_subset(Set1,Set2),
	alsv_subset(Set2,Set1),!.

neqset(Set1,Set2) :-
	\+ eqset(Set1,Set2).

simset(Set1, Set2) :-
	\+ sub(Set1, Set2, Set1).

onesub([],_,[]).	

onesub([L1 to U1|Set1], [L2 to U2],[L1 to NU1, NL2 to U1|R]) :-
	L2 > L1,
	U2 < U1,
	NU1 is L2-1,
	NL2 is U2+1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1], [L2 to U2],[NL1 to U1|R]) :-
	U2 =:= L1,
	U1 > U2,
	NL1 is L1+1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1], [L2 to U2],[L1 to NU1|R]) :-
	U1 =:= L2,
	U2 > U1,
	NU1 is U1-1,
	onesub(Set1,[L2 to U2],R),!.


onesub([L1 to U1|Set1], [L2 to U2],[L1 to NU1|R]) :-
    	L1 < L2,
	L2 < U1,
	U2 >= U1,
	NU1 is L2-1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1], [L2 to U2],[NL1 to U1|R]) :-
	U1 > U2,
	L1 < U2,
	L2 =< L1,
	NL1 is U2+1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1], [L2 to U2],[L2 to NU1|R]) :-
	L2 < L1,
    	L1 < U2,
	U2 =< U1,
    	NU1 is L1-1,
	onesub(Set1,[L2 to U2],R),!.

onesub([L1 to U1|Set1],[L2 to U2], R) :-
	L2 =< L1,
	U1 =< U2,
	onesub(Set1,[L2 to U2],R),!.


onesub([L1 to U1|Set1],[Number],[NL to U1|R]) :-
	L1 \= U1,  
	number(Number),
	Number =:= L1,
	NL is Number+1,
    NL \= U1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[NL|R]) :-
	L1 \= U1,  
	number(Number),
	Number =:= L1,
	NL is Number+1,
    NL =:= U1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[L1 to NU|R]) :-
	L1 \= U1,
	number(Number),
	Number =:= U1,
	NU is Number-1,
    NU \= L1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[NU|R]) :-
	L1 \= U1,
	number(Number),
	Number =:= U1,
	NU is Number-1,
    NU =:= L1,
	onesub(Set1, [Number],R),!.


onesub([L1 to U1|Set1],[Number],[L1 to NU, NL to U1|R]) :-
	L1 \= U1,
	number(Number),
	math_member(Number, [L1 to U1]),
	NU is Number-1,
	NL is Number+1,
    NU \= L1,
    NL \= U1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[NU, NL to U1|R]) :-
	L1 \= U1,
	number(Number),
	math_member(Number, [L1 to U1]),
	NU is Number-1,
	NL is Number+1,
    NU =:= L1,
    NL \= U1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[L1 to NU, NL|R]) :-
	L1 \= U1,
	number(Number),
	math_member(Number, [L1 to U1]),
	NU is Number-1,
	NL is Number+1,
    NU \= L1,
    NL =:= U1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[NU, NL|R]) :-
	L1 \= U1,
	number(Number),
	math_member(Number, [L1 to U1]),
	NU is Number-1,
	NL is Number+1,
    NU =:= L1,
    NL =:= U1,
	onesub(Set1, [Number],R),!.

onesub([L1 to U1|Set1],[Number],[L1 to U1|R]) :-
	L1 \= U1,
	number(Number),
	\+math_member(Number, [L1 to U1]),
	onesub(Set1, [Number],R),!.


onesub([Number|Set1],[L2 to U2],[Number|R]) :-
	number(Number),
	\+ math_member(Number, [L2 to U2]),
	onesub(Set1, [L2 to U2],R),!.

onesub([Number1|Set1],[Number2],[Number1|R]) :-
	number(Number1),
	number(Number2),
	Number1 =\= Number2,	
	onesub(Set1, [Number2],R),!.

onesub([Element1|Set1],[Element2],[Element1|R]) :-
	\+number(Element1),
	\+number(Element2),
	Element1 \= Element2, 
	onesub(Set1,[Element2],R),!.

onesub( [_| Set1],E,R) :-
	onesub(Set1,E,R),!.

sub(R,[],R).

sub(Set1,[H|Set2],R) :-
	onesub(Set1, [H], RR),
	sub(RR,Set2, R),!.

alsv_intersect_all([X|[]],[X|[]]).
alsv_intersect_all([H|RestSet], Result) :-
    alsv_intersect_all(RestSet, RestResult), 
    flatten(H,FH),
    flatten(RestResult,FRestResult),
    alsv_intersect(FH, FRestResult, Result).
	
alsv_union(Set1, Set2, Res) :-
 	merge(Set1, Set2,DRes),
	list_to_set(DRes, Res),!.

alsv_intersect(Set1, Set2, Res) :-
	sub(Set1,Set2,R),
	sub(Set1,R,Res),!.    

alsv_except(Set1, Set2, Res) :-
	sub(Set1, Set2, Res),!.

alsv_completment(Att, Set1, Res) :-
	alsv_domain(Att, Domain, _),
	alsv_attr_class(Att, general),
	sub(Domain, Set1, Res),!.


