:- ensure_loaded('config.pl').
:- ensure_loaded('util.pl').




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% COMPLETENESS                                                             %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- MISSING RULES ------------------------------------------------------------%
%------------------------------------------------------------------------------%	

vcomplete(Schema) :-
  vgenid(_),		% Non-backtrackable predicate
  xtable(Schema, _, Attrs, _),
  vget_states(Schema,Attrs,UStates,CStates), 
  vsplit2details(CStates,DCStates),
  vsplit2details(UStates,DUStates),
  merge(DUStates,DCStates,States), 
  vsplit_intersections(States,DisjointStates),
  xschm(Schema: InputAttributes ==> _),
  vget_rules_states(Schema,CoveredStates),
  dive(Schema,CoveredStates, InputAttributes, DisjointStates, []).
  
vcomplete(_).
 
  
 
 dive(Schema,[], [],_,State) :- 
   vcover_helper(Schema, State),
   fail.
  
 dive(Schema,[],[AttName|Input], DisjointStates, StatePart) :-
   member([AttName,Values], DisjointStates),
   merge([[AttName,Values]], StatePart, StateResult),
   dive(Schema,[],Input,DisjointStates, StateResult).
  
 dive(Schema,CoveredStates,[AttName|Input], DisjointStates, StatePart) :-
   CoveredStates \= [],
   member([AttName,Values], DisjointStates),
   merge([[AttName,Values]], StatePart, StateResult),
   findall(Rule, 
              ( member([ID,CState],CoveredStates),
			      ((member([AttName,CValues],CState),
			        flatten([CValues],FCValues),
				    flatten([Values],FValues),
			        alsv_subset(FValues,FCValues));
				    (\+member([AttName,_],CState))
				   ),
			    Rule=[ID,CState]
			   ),
		   Rules),
   dive(Schema,Rules,Input,DisjointStates, StateResult).
  
  
vget_states(_,[],[],[]).
vget_states(Schema,[A|ARest], UStates, CStates) :- 
  alsv_normalized_domain(A, Domain,_), 
  vget_covered_states(Schema,A,States), %combinations of existing states
  sub(Domain, States,UState1),
  vget_states(Schema,ARest, UStates2,CStates2),
  merge([[A,States]],CStates2, CStates),
  merge([[A,UState1]],UStates2,UStates).

vsplit2details([],[]).
vsplit2details([[Att,Val]|States],Result) :-
  vsplit2details_helper(Att,Val, SingleRes),
  vsplit2details(States,OtherRes),
  merge(SingleRes,OtherRes,Result).


vsplit2details_helper(_,[],[]).
vsplit2details_helper(Att,[V|Vals],[[Att,[V]]|SingleRes]) :-
  vsplit2details_helper(Att,Vals,SingleRes).

vsplit_intersections([],[]).
vsplit_intersections([[A,V]|States], DisjointStates) :-
  findall(Vs,member([A,Vs],[[A,V]|States]),List),
  flatten(List,FList),
  vsplit_intersections_helper(FList, DisjointList), 
  findall([A,[DV]],member(DV,DisjointList),DisStates1),
  findall([AA,VV],(member([AA,VV],States), AA \= A), RestStates),
  vsplit_intersections(RestStates,DisStates2),
  merge(DisStates1,DisStates2,DisjointStates).


vsplit_intersections_helper([],[]).
vsplit_intersections_helper([V|List], DisjointList) :-
  alsv_intersect([V],List, R), 
  R \= [],
  sub([V|List],R, Result), 
  merge(R,Result, NewList),
  vsplit_intersections_helper(NewList,DisjointList),!.
  

vsplit_intersections_helper([V|NewList], [V|DisjointList]) :-
  vsplit_intersections_helper(NewList,DisjointList),!.
  
  
vcover_helper(Schema, StateSet) :- 
  vcover(Schema, StateSet, Res),
  vconcat(['<UncoveredState>', Res ,'</UncoveredState>'],RString), 
  report_file(File),
  vreport(1,[RString],File).


vcover(_,[],'') .
vcover(Schema, [[_, []]|R],Res) :-
  vcover(Schema,R,Res).
vcover(Schema, [[Attr,State]|R],Res) :-
  vcover(Schema, R, InnerRes),
  %scale_down([[Attr,State]],[[_,[ScaledNormState]]]),
  get_explicit_value(Attr,State,ScaledState),
  term_to_atom(ScaledState,AState),
  concat_atom(['<Attribute ID="',Attr,'"> <Value>"',AState,'"</Value></Attribute>',InnerRes],Res),!.


%------------------------------------------------------------------------------%
%--- CONSISTANCE -------------------------------------------------------%
%------------------------------------------------------------------------------%	


%------------------------------------------------------------------------------%
%--- UNUSED ATTRIBUTES -------------------------------------------------------%
%------------------------------------------------------------------------------%	
vunusedatts :-
  findall(Att, (xattr(A),member(name:Att, A)), AllAtts),
  findall(SchmAtt, (xschm(_ : A ==> B),merge(A,B,SchmAtt)),SchmAtts),
  flatten(SchmAtts, FlatSchmAtts),
  subtract(AllAtts, FlatSchmAtts, Rest), !, 
  (
    (Rest = [],!);
    ( term_to_atom(Rest, AtomRest),
      vconcat(['<FreeAttribute ID="', AtomRest ,'"/>'],RString), 
      report_file(File),
      vreport(1,[RString],File))
  ).

   
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% DETERMINISM                                                              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- CONTRADICTION ------------------------------------------------------------%
%------------------------------------------------------------------------------%
vambiguity(Schema) :- 
  vgenid(_), % Non-backtrackable predicate
  vget_rules_states(Schema, RuleStates), 
  ambig(Schema, RuleStates),!.

vambiguity(_). 
  
ambig(_,[]).
ambig(Schema,[[ID, CStates]|Rest])  :-
  ambig_helper(Schema, ID, CStates, Rest, Conflicts),  
  report_conflicts(Schema,Conflicts),
  ambig(Schema,Rest),!.

ambig_helper(_,_,_,[],[]).
ambig_helper(Schema, ID1, SingleState, [[ID2, CStates]|AllRest],[[ID1, ID2, State]|C]) :-
  ambig_helper_helper(Schema, ID1, ID2, SingleState, CStates, State),
  xrule(Schema/ID1: _==>Dec1),
  xrule(Schema/ID2: _==>Dec2),
  ID1 \= ID2,
  conflict(Dec1,Dec2),  
  ambig_helper(Schema, ID1, SingleState, AllRest, C),!.

ambig_helper(Schema, ID1, SingleState, [_|AllRest],C) :-
  ambig_helper(Schema, ID1, SingleState, AllRest, C),!.
  

ambig_helper_helper(_, _, _, [],_,[]).
ambig_helper_helper(Schema, ID1, ID2, [[Att,V1]|StateRest], CStates, [[Att,Confl]|ConflRest]) :-
  member([Att, V2],CStates),
  flatten([V1],FV1),
  flatten([V2], FV2), 
  alsv_intersect(FV1,FV2,Confl), 
  Confl \= [],
  ambig_helper_helper(Schema, ID1, ID2, StateRest, CStates, ConflRest).
 
report_conflicts(_,[]).
report_conflicts(Schema,[[ID1, ID2, State]|Rest]) :- 
  %scale_down(State,ScaledState), 
  report_conflicts_helper(Schema, State, Res),
  vconcat(['<Ambiguous RuleID1="',ID1,'" RuleID2="',ID2,'">',Res,'</Ambiguous>'],RString), 
  report_file(File),
  vreport(0,[RString],File),
  report_conflicts(Schema,Rest),!.
  
report_conflicts_helper(_,[],'').
report_conflicts_helper(Schema, [[_, []]|R],Res) :-
  report_conflicts_helper(Schema,R,Res).
report_conflicts_helper(Schema, [[Attr,State]|R],Res) :- 
  report_conflicts_helper(Schema, R, InnerRes),
  %scale_down([[Attr,State]],[[_,[ScaledNormState]]]),
  get_explicit_value(Attr,State,ScaledState),
  term_to_atom(ScaledState,AState),
  concat_atom(['<Attribute ID="',Attr,'"> <Value>"',AState,'"</Value></Attribute>',InnerRes],Res),!.
  
conflict(Decs1, Decs2):-
  split_decision(Decs1,[D1,_,_,_]), 
  split_decision(Decs2,[D2,_,_,_]),
  conflict_dec_attributes(D1,D2).
  
conflict_dec_attributes([],[]) :- !.
conflict_dec_attributes([D1|Ds1],[D2|Ds2]) :-  
  conflict_item(D1,D2),
  conflict_dec_attributes(Ds1,Ds2).  

conflict_item(D1,D2) :-
  retractall(v(_)),
  assert(v(D2)),
  invalid(D1).
  
%------------------------------------------------------------------------------%
%--- INNER CONTRADICTION ------------------------------------------------------%
%------------------------------------------------------------------------------%
vcontradict(Schema) :-
  vgenid(_), % Non-backtrackable predicate
  vget_rules_states(Schema, RuleStates),
  member(RuleBody, RuleStates),
  RuleBody = [RuleID, CondBody],
  member(Cond, CondBody),
  Cond = [AttName, []],
  vconcat(['<Conflict RuleID="',RuleID,'" AttID="',AttName, '"/>'],RString),
  report_file(File),
  vreport(0,[RString],File),
  fail.

vcontradict(Schema) :-
  xrule(Schema/RuleID: Conds ==> Concls),
  member(Cond, Conds),
  member(Concl, Concls),
  Concl =.. [_,AttName,_],
  Cond =.. [_,AttName,_],
  vconcat(['<Warning RuleID="',RuleID,'" AttID="',AttName, '"/>'],RString),
  report_file(File),
  vreport(0,[RString],File),
  fail.
  
vcontradict(_).

vcontradict_conddec(Schema) :-
  xschm(Schema : Cond ==> Dec),
  intersection(Cond,Dec,Result),
  Result \= [],
  %list all rules from schema that have teh att in cond part
  !.
  
%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
%------------------------------------------------------------------------------%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% MIN. NUMBER OF RULES                                                     %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------------%
%--- SUBSUMPTION --------------------------------------------------------------%
%------------------------------------------------------------------------------%
vsubsume(Schema) :- 
  vgenid(_), % Non-backtrackable predicate
  vget_rules_states(Schema,FinalStates),
  vget_rules_decisions(Schema,FinalDecisions),
  member([RID1,State1],FinalStates),
  member([RID2,State2],FinalStates),
  RID1 \= RID2,
  member([RID1,Dec1],FinalDecisions),
  member([RID2,Dec2],FinalDecisions),
  subsumed_conditions(State1, State2),
  subsumed_decisions(Dec1, Dec2),
  report_subsumptions(Schema,RID1,RID2),
  fail.

vsubsume(_).

subsumed_conditions([],_).
subsumed_conditions([[Att, S1]|State1], State2) :-
  member([Att, S2], State2), 
  flatten(S1,FS1),
  flatten(S2,FS2),
  alsv_subset(FS2,FS1),
  subsumed_conditions(State1,State2).
 
  
subsumed_decisions([],_).
subsumed_decisions([[Att,D1]|Dec1], Dec2) :-
  member([Att,D2],Dec2),
  D1=D2,
  subsumed_decisions(Dec1,Dec2).
  

report_subsumptions(_, RID1, RID2) :-
  vconcat(['<Subsume RuleID1="',RID1,'" RuleID2="',RID2,'"/>'],RString), 
  report_file(File),
  vreport(0,[RString],File).
    

%------------------------------------------------------------------------------%
%--- REDUCTION ----------------------------------------------------------------%
%------------------------------------------------------------------------------%
vreduce(Schema):-
  retractall(pair(_,_)),
  vgenid(_), % Non-backtrackable predicate
  vget_rules_states(Schema,FinalStates),
  vget_rules_decisions(Schema,FinalDecisions),
  member([RID1,_],FinalStates),
  member([RID2,_],FinalStates),
  RID1 \= RID2,
  \+pair(RID2,RID1),
  member([RID1,Dec1],FinalDecisions),
  member([RID2,Dec2],FinalDecisions),
  subtract(Dec1,Dec2,[]),
  assert(pair(RID1,RID2)),
  report_reduction(Schema, RID1,RID2),
  fail.

vreduce(Schema) :-
  retractall(pair(_,_)),
  xrule(Schema/RID1: Conds1 ==> Concls1),
  xrule(Schema/RID2: Conds2 ==> Concls2),
  RID1 \= RID2,
  \+pair(RID2,RID1),
  list_to_ord_set(Conds1,Cnd),
  list_to_ord_set(Conds2,Cnd),
  list_to_ord_set(Concls1,Cnc),
  list_to_ord_set(Concls2,Cnc),
  assert(pair(RID1,RID2)),
  vconcat(['<Equivalent RuleID1="',RID1,'" RuleID2="',RID2,'"/>'],RString), 
  report_file(File),
  vreport(0,[RString],File),
  fail.

vreduce(_) :- 
  retractall(pair(_,_)).

report_reduction(_, RID1, RID2) :-
  vconcat(['<Join RuleID1="',RID1,'" RuleID2="',RID2,'"/>'],RString),
  report_file(File), 
  vreport(0,[RString],File).
  
  
  
